print("import: 'nltk'")
import nltk

print("import: 'nltk.app'")
import nltk.app

print("import: 'nltk.ccg'")
import nltk.ccg

print("import: 'nltk.chat'")
import nltk.chat

print("import: 'nltk.chunk'")
import nltk.chunk

print("import: 'nltk.classify'")
import nltk.classify

print("import: 'nltk.cluster'")
import nltk.cluster

print("import: 'nltk.corpus'")
import nltk.corpus

print("import: 'nltk.corpus.reader'")
import nltk.corpus.reader

print("import: 'nltk.draw'")
import nltk.draw

print("import: 'nltk.inference'")
import nltk.inference

print("import: 'nltk.lm'")
import nltk.lm

print("import: 'nltk.metrics'")
import nltk.metrics

print("import: 'nltk.misc'")
import nltk.misc

print("import: 'nltk.parse'")
import nltk.parse

print("import: 'nltk.sem'")
import nltk.sem

print("import: 'nltk.sentiment'")
import nltk.sentiment

print("import: 'nltk.stem'")
import nltk.stem

print("import: 'nltk.tag'")
import nltk.tag

print("import: 'nltk.tbl'")
import nltk.tbl

print("import: 'nltk.test'")
import nltk.test

print("import: 'nltk.test.unit'")
import nltk.test.unit

print("import: 'nltk.test.unit.translate'")
import nltk.test.unit.translate

print("import: 'nltk.tokenize'")
import nltk.tokenize

print("import: 'nltk.translate'")
import nltk.translate

print("import: 'nltk.twitter'")
import nltk.twitter

